/*
 *  linux/fs/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */
/*
 *  $Id: namei.c,v 1.7 2005/01/26 16:33:37 cwright Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"
#include <linux/quotaops.h>
#include <linux/dnotify.h>

/* This file contains functions lifted from Linux/fs/namei.c. The only reason
 * we have it is to support vfs_unlink_nozombie. */


/*	Check whether we can create an object with dentry child in directory
 *  dir.
 *  1. We can't do it if child already exists (open has special treatment for
 *     this case, but since we are inlined it's OK)
 *  2. We can't do it if dir is read-only (done in permission())
 *  3. We should have write and exec permissions on dir
 *  4. We can't do it if dir is immutable (done in permission())
 */
static inline int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
may_create(struct inode *dir, struct dentry *child)
#else
may_create(struct inode *dir, struct dentry *child, struct nameidata *nd)
#endif
{
	if (child->d_inode)
		return -EEXIST;
	if (IS_DEADDIR(dir))
		return -ENOENT;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	return permission(dir,MAY_WRITE | MAY_EXEC);
#else
	return permission(dir,MAY_WRITE | MAY_EXEC, nd);
#endif
}

//DQ: vfs_create now also takes a nameidata struct we might want to make this one take one also.
/* Lifted from Linux/fs/namei.c:960, but will not down dir->i_zombie. */
int vfs_create_nozombie(struct inode *dir, struct dentry *dentry, int mode)
{
	int error;

	mode &= S_IALLUGO;
	mode |= S_IFREG;

	/* REMOVED: down(&dir->i_zombie); */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	error = may_create(dir, dentry);
#else
	//DQ: the null gets changed to a nameidata struct if we choose to add it here.
	error = may_create(dir, dentry,NULL);
#endif
	if (error)
		goto exit_lock;

	error = -EACCES;	/* shouldn't it be ENOSYS? */
	if (!dir->i_op || !dir->i_op->create)
		goto exit_lock;

	DQUOT_INIT(dir);
	lock_kernel();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	error = dir->i_op->create(dir, dentry, mode);
#else
	error = dir->i_op->create(dir, dentry, mode, NULL);
#endif
	unlock_kernel();
exit_lock:
	/* REMOVED: up(&dir->i_zombie); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	if (!error)
		inode_dir_notify(dir, DN_CREATE);
#endif
	return error;
}
